@@ -135,10 +135,17 @@ struct t3_cookie {
 	uint8_t mc_block[BLOCK_SIZE];
 };
 
-static void t3_cookie_release(struct t3_cookie *cookie)
+static int t3_cookie_release(int err, void *data)
 {
+	struct t3_cookie *cookie = data;
+
+	DBG("%p", cookie);
+
 	if (cookie == NULL)
-		return;
+		return err;
+
+	if (cookie->cb != NULL)
+		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
 
 	if (cookie->ndef != NULL)
 		g_free(cookie->ndef->data);
@@ -146,6 +153,8 @@ static void t3_cookie_release(struct t3_cookie *cookie)
 	g_free(cookie->ndef);
 	g_free(cookie);
 	cookie = NULL;
+
+	return err;
 }
 
 /* common: Initialize structure to write block */
@@ -170,9 +179,9 @@ static void prepare_write_block(uint8_t *UID, struct type3_cmd *cmd,
 /* common: Initialize structure to read block */
 static void prepare_read_block(uint8_t cur_block,
 				uint8_t *UID,
-				struct type3_cmd *cmd )
+				struct type3_cmd *cmd)
 {
-	cmd->cmd	 = CMD_READ_WO_ENCRYPT;		/* command */
+	cmd->cmd = CMD_READ_WO_ENCRYPT;			/* command */
 	memcpy(cmd->data, UID, LEN_ID);			/* IDm */
 
 	cmd->data[LEN_ID] = 1;				/* number of service */
@@ -189,22 +198,20 @@ static void prepare_read_block(uint8_t cur_block,
 /* common: Simple checks on received frame */
 static int check_recv_frame(uint8_t *resp, uint8_t reply_code)
 {
-	int err = 0;
-
 	if (resp[OFS_NFC_STATUS] != 0) {
-		DBG("NFC Command failed: 0x%x",resp[OFS_NFC_STATUS]);
-		err = -EIO;
+		DBG("NFC Command failed: 0x%x", resp[OFS_NFC_STATUS]);
+		return -EIO;
 	}
 
-	if (resp[OFS_CMD_RESP] != reply_code ) {
+	if (resp[OFS_CMD_RESP] != reply_code) {
 		DBG("Felica cmd failed: 0x%x", resp[OFS_CMD_RESP]);
-		err = -EIO;
+		return -EIO;
 	}
 
-	return err;
+	return 0;
 }
 
-static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
+static int data_recv(uint8_t *resp, int length, void *data)
 {
 	struct type3_tag *tag = data;
 	struct type3_cmd cmd;
@@ -222,11 +229,11 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	nfc_data = near_tag_get_data(tag->tag, &data_length);
-	length_read = length - OFS_READ_DATA  ;
+	length_read = length - OFS_READ_DATA;
 	current_length = tag->current_block * BLOCK_SIZE;
 	if (current_length + (length - OFS_READ_DATA) > data_length)
 		length_read = data_length - current_length;
@@ -239,7 +246,7 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 		tag->current_block = 0;
 
 		DBG("Done reading %zd bytes at %p", data_length, nfc_data);
-		records = near_ndef_parse(nfc_data, data_length);
+		records = near_ndef_parse_msg(nfc_data, data_length);
 		near_tag_add_records(tag->tag, records, tag->cb, 0);
 
 		g_free(tag);
@@ -252,16 +259,17 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 	tag->current_block += read_blocks;
 
 	prepare_read_block(DATA_BLOCK_START + tag->current_block,
-				tag->IDm, &cmd );
+				tag->IDm, &cmd);
+
+	err = near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+				data_recv, tag, NULL);
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd, cmd.len,
-			nfctype3_data_recv, tag);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
+out_err:
 	if (err < 0 && tag->cb)
 		tag->cb(adapter_idx, target_idx, err);
 
@@ -270,7 +278,7 @@ out:
 	return err;
 }
 
-static int nfctype3_data_read(struct type3_tag *tag)
+static int data_read(struct type3_tag *tag)
 {
 	struct type3_cmd cmd;
 	uint32_t adapter_idx;
@@ -286,7 +294,7 @@ static int nfctype3_data_read(struct type3_tag *tag)
 
 	return near_adapter_send(adapter_idx,
 					(uint8_t *) &cmd, cmd.len,
-					nfctype3_data_recv, tag);
+					data_recv, tag, NULL);
 }
 
 /* Read block 0 to retrieve the data length */
@@ -296,23 +304,23 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	int err = 0;
 	struct near_tag *tag;
 	struct type3_tag *t3_tag = NULL;
-	uint32_t  ndef_data_length;
+	uint32_t ndef_data_length;
 
 	DBG("%d", length);
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (resp[OFS_READ_FLAG] != 0) {
 		DBG("Status 0x%x", resp[OFS_READ_FLAG]);
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	/* Block 0:[11 - 13]: length is a 3 bytes value */
@@ -325,12 +333,12 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx,
 					NULL, ndef_data_length);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
 	/* Block 0:[10]: RW Flag. 1 for RW */
@@ -342,10 +350,10 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	t3_tag = g_try_malloc0(sizeof(struct type3_tag));
 	if (t3_tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
-	memcpy(t3_tag->IDm, cookie->IDm , LEN_ID);
+	memcpy(t3_tag->IDm, cookie->IDm, LEN_ID);
 
 	near_tag_set_idm(tag, cookie->IDm, LEN_ID);
 	near_tag_set_attr_block(tag, resp + OFS_READ_DATA, BLOCK_SIZE);
@@ -356,20 +364,20 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	t3_tag->cb = cookie->cb;
 	t3_tag->tag = tag;
 
-	err = nfctype3_data_read(t3_tag);
+	err = data_read(t3_tag);
 
-out:
-	if (err < 0) {
-		if (cookie->cb)
-			cookie->cb(cookie->adapter_idx, cookie->target_idx,
-									err);
+	/*
+	 * As reading isn't complete,
+	 * callback shouldn't be called while freeing the cookie
+	 */
+	if (err == 0)
+		cookie->cb = NULL;
 
+out_err:
+	if (err < 0)
 		g_free(t3_tag);
-	}
 
-	t3_cookie_release(cookie);
-
-	return err;
+	return t3_cookie_release(err, cookie);
 }
 
 static int poll_ndef_system_code(uint8_t *resp, int length, void *data)
@@ -378,34 +386,28 @@ static int poll_ndef_system_code(uint8_t *resp, int length, void *data)
 	int err = 0;
 	struct type3_cmd cmd;
 
-	DBG(" length: %d", length);
+	DBG("length: %d", length);
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_POLL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	prepare_read_block(META_BLOCK_START, cookie->IDm, &cmd);
 
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-				cmd.len, nfctype3_recv_block_0, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, nfctype3_recv_block_0, cookie, NULL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx,
-				cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int check_sys_op_in_mc_block(uint8_t *resp, int length, void *data)
@@ -419,12 +421,12 @@ static int check_sys_op_in_mc_block(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (resp[SYSTEM_OPTION_OFFSET] == 0x00) {
 		DBG("Blank tag detected");
@@ -433,26 +435,20 @@ static int check_sys_op_in_mc_block(uint8_t *resp, int length, void *data)
 					cookie->target_idx,
 					NULL, 1 /* dummy length */);
 		if (err < 0)
-			goto out;
+			goto out_err;
 
 		tag = near_tag_get_tag(cookie->adapter_idx,
 					cookie->target_idx);
 		if (tag == NULL) {
 			err = -ENOMEM;
-			goto out;
+		goto out_err;
 		}
 
 		near_tag_set_idm(tag, cookie->IDm, LEN_ID);
 		near_tag_set_ic_type(tag, cookie->ic_type);
 		near_tag_set_blank(tag, TRUE);
 
-		if (cookie->cb)
-			cookie->cb(cookie->adapter_idx,
-					cookie->target_idx, 0);
-
-		t3_cookie_release(cookie);
-
-		return 0;
+		return t3_cookie_release(0, cookie);
 	} else {
 		/* CMD POLL */
 		cmd.cmd	 = CMD_POLL;	/* POLL command */
@@ -463,22 +459,17 @@ static int check_sys_op_in_mc_block(uint8_t *resp, int length, void *data)
 		/* data len + 2 bytes */
 		cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
 
-		err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-			cmd.len , poll_ndef_system_code, cookie);
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+			cmd.len , poll_ndef_system_code, cookie, NULL);
 	}
 
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int receive_system_code(uint8_t *resp, int length, void *data)
@@ -488,28 +479,28 @@ static int receive_system_code(uint8_t *resp, int length, void *data)
 	struct type3_cmd cmd;
 	uint16_t system_code;
 
-	DBG(" length: %d", length);
+	DBG("length: %d", length);
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_POLL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	cookie->ic_type = resp[IC_TYPE_OFFSET];
 	memcpy(cookie->IDm, resp + OFS_IDM, LEN_ID);
-	system_code = ((uint16_t)(resp[length - 2])) << 8;
+	system_code = ((uint16_t) (resp[length - 2])) << 8;
 	system_code |= resp[length - 1];
 
 	switch (resp[IC_TYPE_OFFSET]) {
 	case FELICA_LITE_IC_TYPE:
 		prepare_read_block(FELICA_LITE_MC_BLOCK, cookie->IDm, &cmd);
-		err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
 					cmd.len, check_sys_op_in_mc_block,
-					cookie);
+					cookie, NULL);
 		break;
 
 	case FELICA_LITE_S_IC_TYPE:
@@ -523,22 +514,18 @@ static int receive_system_code(uint8_t *resp, int length, void *data)
 		/* data len + 2 bytes */
 		cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
 
-		err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-					cmd.len, poll_ndef_system_code, cookie);
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+					cmd.len, poll_ndef_system_code, cookie,
+					NULL);
 	}
 
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int nfctype3_read(uint32_t adapter_idx,
@@ -546,7 +533,6 @@ static int nfctype3_read(uint32_t adapter_idx,
 {
 	struct type3_cmd cmd;
 	struct t3_cookie *cookie;
-	int err;
 
 	DBG("");
 
@@ -558,7 +544,7 @@ static int nfctype3_read(uint32_t adapter_idx,
 	cmd.data[3] = 0x00;	/* time slot */
 
 	/* data len + 2 bytes */
-	cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
+	cmd.len = LEN_CMD + LEN_CMD_LEN + 4;
 
 	cookie = g_try_malloc0(sizeof(struct t3_cookie));
 	if (cookie == NULL)
@@ -568,12 +554,9 @@ static int nfctype3_read(uint32_t adapter_idx,
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd,
-				cmd.len , receive_system_code, cookie);
-	if (err < 0)
-		g_free(cookie);
-
-	return err;
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+					receive_system_code, cookie,
+					t3_cookie_release);
 }
 
 static int update_attr_block_cb(uint8_t *resp, int length, void *data)
@@ -585,22 +568,17 @@ static int update_attr_block_cb(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	DBG("Done writing");
 
-out:
-	if (cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int update_attr_block(struct t3_cookie *cookie)
@@ -626,7 +604,8 @@ static int update_attr_block(struct t3_cookie *cookie)
 	prepare_write_block(cookie->IDm, &cmd, 0, cookie->attr);
 
 	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
-						update_attr_block_cb, cookie);
+					update_attr_block_cb, cookie,
+					t3_cookie_release);
 }
 
 static int data_write_resp(uint8_t *resp, int length, void *data)
@@ -640,17 +619,17 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (cookie->ndef->offset >= cookie->ndef->length) {
 		err = update_attr_block(cookie);
 		if (err < 0)
-			goto out;
+			goto out_err;
 
 		return 0;
 	}
@@ -669,20 +648,15 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 	cookie->current_block++;
 	cookie->ndef->offset += BLOCK_SIZE;
 
-	err =  near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd, cmd.len,
-						data_write_resp, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
+						data_write_resp, cookie, NULL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int data_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -700,9 +674,14 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t3_cookie));
+
 	if (cookie == NULL) {
 		err = -ENOMEM;
-		goto out;
+
+		if (cb != NULL)
+			cb(adapter_idx, target_idx, err);
+
+		return err;
 	}
 
 	cookie->adapter_idx = adapter_idx;
@@ -712,26 +691,22 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	cookie->current_block = 0;
 
 	idm = near_tag_get_idm(tag, &len);
-	if (idm == NULL) {
-		err = -EINVAL;
-		goto out;
-	}
+	if (idm == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
 
 	memcpy(cookie->IDm, idm, len);
 
 	attr = near_tag_get_attr_block(tag, &len);
-	if (attr == NULL) {
-		err = -EINVAL;
-		goto out;
-	}
+	if (attr == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
 
 	memcpy(cookie->attr, attr, len);
-	nmaxb = (((uint16_t)(cookie->attr[3])) << 8) | cookie->attr[4];
+	nmaxb = (((uint16_t) (cookie->attr[3])) << 8) | cookie->attr[4];
 
 	if (cookie->ndef->length > (nmaxb * BLOCK_SIZE)) {
 		near_error("not enough space on tag");
-		err = -ENOSPC;
-		goto out;
+
+		return t3_cookie_release(-ENOSPC, cookie);
 	}
 
 	cookie->attr[9] = 0x0F; /* writing data in progress */
@@ -740,24 +715,16 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	for (i = 0; i < 14; i++)
 		checksum += cookie->attr[i];
 
-	cookie->attr[14] = (uint8_t)(checksum >> 8);
-	cookie->attr[15] = (uint8_t)(checksum);
+	cookie->attr[14] = (uint8_t) (checksum >> 8);
+	cookie->attr[15] = (uint8_t) checksum;
 
 	prepare_write_block(cookie->IDm, &cmd, cookie->current_block,
 							cookie->attr);
 	cookie->current_block++;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd, cmd.len,
-						data_write_resp, cookie);
-	if (err < 0)
-		goto out;
-
-	return 0;
-
-out:
-	t3_cookie_release(cookie);
-
-	return err;
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+					data_write_resp, cookie,
+					t3_cookie_release);
 }
 
 static int nfctype3_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -765,22 +732,33 @@ static int nfctype3_write(uint32_t adapter_idx, uint32_t target_idx,
 				near_tag_io_cb cb)
 {
 	struct near_tag *tag;
+	int err;
 
 	DBG("");
 
-	if (ndef == NULL || cb == NULL)
-		return -EINVAL;
+	if (ndef == NULL || cb == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-	if (tag == NULL)
-		return -EINVAL;
+	if (tag == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
-	return data_write(adapter_idx, target_idx, ndef, tag, cb);
+	err = data_write(adapter_idx, target_idx, ndef, tag, cb);
+
+out_err:
+	if (cb != NULL)
+		cb(adapter_idx, target_idx, err);
+
+	return err;
 }
 
-static int nfctype3_check_recv_UID(uint8_t *resp, int length, void *data)
+static int check_presence(uint8_t *resp, int length, void *data)
 {
-	struct t3_cookie *rcv_cookie = data;
+	struct t3_cookie *cookie = data;
 	int err = 0;
 
 	DBG("length %d", length);
@@ -788,13 +766,7 @@ static int nfctype3_check_recv_UID(uint8_t *resp, int length, void *data)
 	if (length < 0)
 		err = -EIO;
 
-	if (rcv_cookie->cb)
-		rcv_cookie->cb(rcv_cookie->adapter_idx,
-				rcv_cookie->target_idx, err);
-
-	t3_cookie_release(rcv_cookie);
-
-	return err;
+	return t3_cookie_release(err, cookie);
 }
 
 static int nfctype3_check_presence(uint32_t adapter_idx,
@@ -802,7 +774,6 @@ static int nfctype3_check_presence(uint32_t adapter_idx,
 {
 	struct type3_cmd cmd;
 	struct t3_cookie *cookie;
-	int err;
 
 	DBG("");
 
@@ -824,17 +795,9 @@ static int nfctype3_check_presence(uint32_t adapter_idx,
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd,
-			cmd.len , nfctype3_check_recv_UID, cookie);
-	if (err < 0)
-		goto out;
-
-	return 0;
-
-out:
-	t3_cookie_release(cookie);
-
-	return err;
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd,
+				cmd.len, check_presence, cookie,
+				t3_cookie_release);
 }
 
 static int format_resp(uint8_t *resp, int length, void *data)
@@ -847,17 +810,17 @@ static int format_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
 	near_tag_set_ro(tag, FALSE);
@@ -867,13 +830,8 @@ static int format_resp(uint8_t *resp, int length, void *data)
 
 	DBG("Formatting is done");
 
-out:
-	if (cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int write_attr_block(uint8_t *resp, int length , void *data)
@@ -887,12 +845,12 @@ static int write_attr_block(uint8_t *resp, int length , void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	cookie->attr[0] = NDEF_MAPPING_VERSION;
 	cookie->attr[1] = MAX_READ_BLOCKS_PER_CHECK;
@@ -918,20 +876,15 @@ static int write_attr_block(uint8_t *resp, int length , void *data)
 	prepare_write_block(cookie->IDm, &cmd, META_BLOCK_START,
 						cookie->attr);
 
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-				cmd.len, format_resp, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, format_resp, cookie, NULL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int write_mc_block(uint8_t *resp, int length, void *data)
@@ -944,17 +897,17 @@ static int write_mc_block(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (resp[OFS_READ_FLAG] != 0) {
 		DBG("Status 0x%x", resp[OFS_READ_FLAG]);
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	memcpy(cookie->mc_block, resp + 14, BLOCK_SIZE);
@@ -965,20 +918,15 @@ static int write_mc_block(uint8_t *resp, int length, void *data)
 	cookie->mc_block[3] = 1;
 	prepare_write_block(cookie->IDm, &cmd, FELICA_LITE_MC_BLOCK,
 				cookie->mc_block);
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-				cmd.len, write_attr_block, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, write_attr_block, cookie, NULL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int nfctype3_format(uint32_t adapter_idx,
@@ -989,7 +937,6 @@ static int nfctype3_format(uint32_t adapter_idx,
 	struct t3_cookie *cookie;
 	uint8_t ic_type;
 	uint8_t *idm, len;
-	int err;
 
 	DBG("");
 
@@ -1011,25 +958,17 @@ static int nfctype3_format(uint32_t adapter_idx,
 	cookie->ic_type = ic_type;
 
 	idm = near_tag_get_idm(tag, &len);
-	if (idm == NULL) {
-		err = -EINVAL;
-		goto out;
-	}
+	if (idm == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
 
 	memcpy(cookie->IDm, idm, len);
 
 	prepare_read_block(FELICA_LITE_MC_BLOCK, cookie->IDm, &cmd);
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-				cmd.len, write_mc_block, cookie);
-	if (err < 0)
-		goto out;
 
-	return 0;
+	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
+					write_mc_block, cookie,
+					t3_cookie_release);
 
-out:
-	t3_cookie_release(cookie);
-
-	return err;
 }
 
 static struct near_tag_driver type1_driver = {
