@@ -40,7 +40,7 @@
 #include <near/ndef.h>
 #include <near/tlv.h>
 
-#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
+#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
 
 #define NFC_STATUS		0
 #define NFC_STATUS_BYTE_LEN	1
@@ -158,6 +158,27 @@ struct t4_cookie {
 	uint16_t memory_size;
 };
 
+static int t4_cookie_release(int err, void *data)
+{
+	struct t4_cookie *cookie = data;
+
+	DBG("%p", cookie);
+
+	if (cookie == NULL)
+		return err;
+
+	if (err < 0 && cookie->cb)
+		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
+
+	if (cookie->ndef)
+		g_free(cookie->ndef->data);
+
+	g_free(cookie->ndef);
+	g_free(cookie);
+
+	return err;
+}
+
 /* ISO functions: This code prepares APDU */
 static int ISO_send_cmd(uint8_t class,
 			uint8_t instruction,
@@ -207,10 +228,9 @@ static int ISO_send_cmd(uint8_t class,
 			cmd->data[cmd_data_length] = 0;
 	}
 
-	err =  near_adapter_send(in_rcv->adapter_idx, (uint8_t *)cmd,
-					total_cmd_length, cb, in_rcv);
-	if (err < 0)
-		g_free(in_rcv);
+	return near_adapter_send(in_rcv->adapter_idx, (uint8_t *) cmd,
+					total_cmd_length, cb, in_rcv,
+					t4_cookie_release);
 
 out_err:
 	/* On exit, clean memory */
@@ -248,8 +268,8 @@ static int ISO_ReadBinary(uint16_t offset, uint8_t readsize,
 	return ISO_send_cmd(
 			0x00,		/* CLA */
 			0xB0,		/* INS: Select file */
-			(uint8_t)((offset & 0xFF00)>>8),
-			(uint8_t)(offset & 0xFF),
+			(uint8_t) ((offset & 0xFF00) >> 8),
+			(uint8_t) (offset & 0xFF),
 			0,		/* no data send */
 			readsize,	/* bytes to read */
 			FALSE,
@@ -265,8 +285,8 @@ static int ISO_Update(uint16_t offset, uint8_t nlen,
 	return ISO_send_cmd(
 			0x00,			/* CLA */
 			0xD6,			/* INS: Select file */
-			(uint8_t)((offset & 0xFF00) >> 8),
-			(uint8_t)(offset & 0xFF),
+			(uint8_t) ((offset & 0xFF00) >> 8),
+			(uint8_t) (offset & 0xFF),
 			data,			/* length of NDEF data */
 			nlen,			/* NLEN + NDEF data */
 			FALSE,
@@ -274,40 +294,22 @@ static int ISO_Update(uint16_t offset, uint8_t nlen,
 			cookie);
 }
 
-static int t4_cookie_release(int err, struct t4_cookie *cookie)
-{
-	if (cookie == NULL)
-		return err;
-
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	if (cookie->ndef)
-		g_free(cookie->ndef->data);
-
-	g_free(cookie->ndef);
-	g_free(cookie);
-
-	return err;
-}
-
 static int data_read_cb(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data ;
 	uint8_t *nfc_data;
 	size_t data_length, length_read, current_length;
 	uint16_t remain_bytes;
-	int err = 0;
 
 	DBG("%d", length);
 
 	if (length < 0)
-		return t4_cookie_release(err, cookie);
+		return t4_cookie_release(length, cookie);
 
 	if (APDU_STATUS(resp + length - 2) != APDU_OK) {
 		DBG("Fail read_cb SW:x%04x", APDU_STATUS(resp + length - 2));
-		err = -EIO;
-		return t4_cookie_release(err, cookie);
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	nfc_data = near_tag_get_data(cookie->tag, &data_length);
@@ -327,61 +329,49 @@ static int data_read_cb(uint8_t *resp, int length, void *data)
 
 		DBG("Done reading");
 
-		records = near_ndef_parse(nfc_data, data_length);
+		records = near_ndef_parse_msg(nfc_data, data_length);
 		near_tag_add_records(cookie->tag, records, cookie->cb, 0);
 
-		err = 0;
-		goto out_err;
+		return t4_cookie_release(0, cookie);
 	}
 
 	cookie->read_data += length ;
 	remain_bytes = (data_length - cookie->read_data);
 
 	if (remain_bytes >= cookie->r_apdu_max_size)
-		err = ISO_ReadBinary(cookie->read_data + 2,
+		return ISO_ReadBinary(cookie->read_data + 2,
 				cookie->r_apdu_max_size, data_read_cb, cookie);
 	else
-		err = ISO_ReadBinary(cookie->read_data + 2,
-				(uint8_t)remain_bytes, data_read_cb, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
+		return ISO_ReadBinary(cookie->read_data + 2,
+				(uint8_t) remain_bytes, data_read_cb, cookie);
 }
 
 static int t4_readbin_NDEF_ID(uint8_t *resp, int length, void *data)
 {
-	struct t4_cookie *cookie = data ;
+	struct t4_cookie *cookie = data;
 	struct near_tag *tag;
-	int err = 0;
+	int err;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	if (APDU_STATUS(resp + length - 2) != APDU_OK) {
 		DBG("Fail SW:x%04x", APDU_STATUS(resp + length - 2));
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	/* Add data to the tag */
 	err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx, NULL,
 				g_ntohs(*((uint16_t *)(resp + NFC_STATUS_BYTE_LEN))));
 	if (err < 0)
-		goto out_err;
+		return t4_cookie_release(err, cookie);
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
-	if (tag == NULL) {
-		err = -ENOMEM;
-		goto out_err;
-	}
+	if (tag == NULL)
+		return t4_cookie_release(-ENOMEM, cookie);
 
 	near_tag_set_max_ndef_size(tag, cookie->max_ndef_size);
 	near_tag_set_c_apdu_max_size(tag, cookie->c_apdu_max_size);
@@ -396,121 +386,93 @@ static int t4_readbin_NDEF_ID(uint8_t *resp, int length, void *data)
 	else
 		near_tag_set_ro(tag, FALSE);
 
-	/* TODO: see how we can get the UID value:
-	 *  near_tag_set_uid(tag, resp + NFC_HEADER_SIZE, 8);
-	 *  */
+	/*
+	 * TODO: see how we can get the UID value:
+	 * near_tag_set_uid(tag, resp + NFC_HEADER_SIZE, 8);
+	 */
 
 	/* Read 1st block */
-	err = ISO_ReadBinary(2, cookie->r_apdu_max_size - 2,
+	return ISO_ReadBinary(2, cookie->r_apdu_max_size - 2,
 			data_read_cb, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 static int t4_select_NDEF_ID(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	int err = 0;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	/* Check for APDU error */
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
 		DBG("Fail SW:x%04x", APDU_STATUS(resp + STATUS_WORD_1));
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	/* Read 0x0f bytes, to grab the NDEF msg length */
-	err = ISO_ReadBinary(0, LEN_ISO_CC_READ_SIZE,
+	return ISO_ReadBinary(0, LEN_ISO_CC_READ_SIZE,
 					t4_readbin_NDEF_ID, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 static int t4_readbin_cc(uint8_t *resp, int length, void *data)
 {
-	struct t4_cookie *cookie = data ;
-	struct type4_cc	*read_cc ;
-	int err = 0;
+	struct t4_cookie *cookie = data;
+	struct type4_cc *read_cc;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	/* Check APDU error ( the two last bytes of the resp) */
 	if (APDU_STATUS(resp + length - 2) != APDU_OK) {
 		DBG("Fail SW:x%04x", APDU_STATUS(resp + length - 2));
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	/* -2 for status word and -1 is for NFC first byte... */
 	read_cc = g_try_malloc0(length - 2 - NFC_STATUS_BYTE_LEN);
 	if (read_cc == NULL) {
 		DBG("Mem alloc failed");
-		err = -ENOMEM;
-		goto out_err;
+
+		return t4_cookie_release(-ENOMEM, cookie);
 	}
 
 	memcpy(read_cc, &resp[1], length - 2 - NFC_STATUS_BYTE_LEN) ;
 
 	cookie->r_apdu_max_size = g_ntohs(read_cc->max_R_apdu_data_size) -
-			APDU_HEADER_LEN ;
+			APDU_HEADER_LEN;
 	cookie->c_apdu_max_size = g_ntohs(read_cc->max_C_apdu_data_size);
 	cookie->max_ndef_size = g_ntohs(read_cc->tlv_fc.max_ndef_size);
 
-	/* TODO 5.1.1 :TLV blocks can be zero, one or more...  */
-	/* TODO 5.1.2 :Must ignore proprietary blocks (x05)...  */
-	if (read_cc->tlv_fc.tag  != 0x4) {
-		DBG("NDEF File Control tag not found !") ;
-		err = -EINVAL ;
-		goto out_err ;
+	/* TODO 5.1.1: TLV blocks can be zero, one or more... */
+	/* TODO 5.1.2: Must ignore proprietary blocks (x05)... */
+	if (read_cc->tlv_fc.tag != 0x4) {
+		DBG("NDEF File Control tag not found");
+
+		return t4_cookie_release(-EINVAL, cookie);
 	}
 
 	/* save rw conditions */
 	cookie->write_access = read_cc->tlv_fc.write_access;
 
-	err = ISO_Select((uint8_t *)&read_cc->tlv_fc.file_id,
+	return ISO_Select((uint8_t *) &read_cc->tlv_fc.file_id,
 			LEN_ISO_CC_FILEID, 0, t4_select_NDEF_ID, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 static int t4_select_cc(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	int err = 0;
+	int err;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	/* Check for APDU error */
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
@@ -520,14 +482,12 @@ static int t4_select_cc(uint8_t *resp, int length, void *data)
 		err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx,
 					NULL, 1 /* dummy length */);
 		if (err < 0)
-			goto out_err;
+			return t4_cookie_release(err, cookie);
 
 		cookie->tag = near_tag_get_tag(cookie->adapter_idx,
 						cookie->target_idx);
-		if (cookie->tag == NULL) {
-			err = -ENOMEM;
-			goto out_err;
-		}
+		if (cookie->tag == NULL)
+			return t4_cookie_release(-ENOMEM, cookie);
 
 		near_tag_set_blank(cookie->tag, TRUE);
 
@@ -537,128 +497,84 @@ static int t4_select_cc(uint8_t *resp, int length, void *data)
 		return t4_cookie_release(0, cookie);
 	}
 
-	err = ISO_ReadBinary(0, LEN_ISO_CC_READ_SIZE, t4_readbin_cc, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
+	return ISO_ReadBinary(0, LEN_ISO_CC_READ_SIZE, t4_readbin_cc, cookie);
 }
 
 static int t4_select_file_by_name_v1(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	int err = 0 ;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	/* Check for APDU error */
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
 		DBG("V1 Fail SW:x%04x", APDU_STATUS(resp + STATUS_WORD_1));
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
-	if (resp[NFC_STATUS] != 0) {
-		err = -EIO;
-		goto out_err;
-	}
+	if (resp[NFC_STATUS] != 0)
+		return t4_cookie_release(-EIO, cookie);
 
 	/* Jump to select phase */
-	err = ISO_Select(iso_cc_fileid, LEN_ISO_CC_FILEID, 0,
+	return ISO_Select(iso_cc_fileid, LEN_ISO_CC_FILEID, 0,
 				t4_select_cc, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 static int t4_select_file_by_name_v2(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	int err = 0 ;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		t4_cookie_release(length, cookie);
 
 	/* Check for APDU error - Not found */
 	if (APDU_STATUS(resp + STATUS_WORD_1) == APDU_NOT_FOUND) {
 		DBG("Fallback to V1");
-		err = ISO_Select(iso_appname_v1, ARRAY_SIZE(iso_appname_v1),
-				0x4, t4_select_file_by_name_v1, cookie);
-		if (err < 0)
-			goto out_err;
 
-		return err;
+		return ISO_Select(iso_appname_v1, ARRAY_SIZE(iso_appname_v1),
+					0x4, t4_select_file_by_name_v1, cookie);
 	}
 
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
 		DBG("V2 Fail SW:x%04x", APDU_STATUS(resp + STATUS_WORD_1));
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
-	if (resp[NFC_STATUS] != 0) {
-		err = -EIO;
-		goto out_err;
-	}
+	if (resp[NFC_STATUS] != 0)
+		return t4_cookie_release(-EIO, cookie);
 
 	/* Jump to select phase */
-	err = ISO_Select(iso_cc_fileid, LEN_ISO_CC_FILEID, 0,
-			t4_select_cc, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
+	return ISO_Select(iso_cc_fileid, LEN_ISO_CC_FILEID, 0, t4_select_cc,
+									cookie);
 }
 
 static int nfctype4_read(uint32_t adapter_idx,
 		uint32_t target_idx, near_tag_io_cb cb)
 {
 	struct t4_cookie *cookie;
-	int err = 0;
 
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
-	if (cookie == NULL) {
-		err = -ENOMEM;
-		goto out_err;
-	}
+	if (cookie == NULL)
+		return -ENOMEM;
 
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 	cookie->tag = NULL;
-	cookie->read_data = 0;;
+	cookie->read_data = 0;
 
 	/* Check for V2 type 4 tag */
-	err = ISO_Select(iso_appname_v2, ARRAY_SIZE(iso_appname_v2),
-			0x4, t4_select_file_by_name_v2, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
+	return ISO_Select(iso_appname_v2, ARRAY_SIZE(iso_appname_v2),
+				0x4, t4_select_file_by_name_v2, cookie);
 }
 
 static int data_write_cb(uint8_t *resp, int length, void *data)
@@ -669,14 +585,13 @@ static int data_write_cb(uint8_t *resp, int length, void *data)
 	DBG("%d", length);
 
 	if (length < 0)
-		return t4_cookie_release(err, cookie);
+		return t4_cookie_release(length, cookie);
 
 	if (APDU_STATUS(resp + length - 2) != APDU_OK) {
 		near_error("write failed SWx%04x",
 				APDU_STATUS(resp + length - 2));
-		err = -EIO;
 
-		return t4_cookie_release(err, cookie);
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	if (cookie->ndef->offset >= cookie->ndef->length) {
@@ -718,9 +633,14 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	int err;
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
+
 	if (cookie == NULL) {
 		err = -ENOMEM;
-		goto out;
+
+		if (cb != NULL)
+			cb(adapter_idx, target_idx, err);
+
+		return err;
 	}
 
 	cookie->adapter_idx = adapter_idx;
@@ -734,8 +654,8 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 
 	if (cookie->max_ndef_size < cookie->ndef->length) {
 		near_error("not enough space on tag to write data");
-		err = -ENOMEM;
-		goto out;
+
+		return t4_cookie_release(-ENOMEM, cookie);
 	}
 
 	if ((cookie->ndef->length - cookie->ndef->offset) >
@@ -754,31 +674,40 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	}
 
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	t4_cookie_release(err, cookie);
-
-	return err;
+out_err:
+	return t4_cookie_release(err, cookie);
 }
 
 static int nfctype4_write(uint32_t adapter_idx, uint32_t target_idx,
 			struct near_ndef_message *ndef, near_tag_io_cb cb)
 {
 	struct near_tag *tag;
+	int err;
 
 	DBG("");
 
-	if (ndef == NULL || cb == NULL)
-		return -EINVAL;
+	if (ndef == NULL || cb == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-	if (tag == NULL)
-		return -EINVAL;
+	if (tag == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
-	return data_write(adapter_idx, target_idx, ndef, tag, cb);
+	err = data_write(adapter_idx, target_idx, ndef, tag, cb);
+
+out_err:
+	if (cb != NULL)
+		cb(adapter_idx, target_idx, err);
+
+	return err;
 }
 
 static int check_presence(uint8_t *resp, int length, void *data)
@@ -802,51 +731,38 @@ static int nfctype4_check_presence(uint32_t adapter_idx,
 		uint32_t target_idx, near_tag_io_cb cb)
 {
 	struct t4_cookie *cookie;
-	int err = 0;
 
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t4_cookie));
-	if (cookie == NULL) {
-		err = -ENOMEM;
-		goto out_err;
-	}
+	if (cookie == NULL)
+		return -ENOMEM;
 
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 	cookie->tag = NULL;
-	cookie->read_data = 0;;
+	cookie->read_data = 0;
 
 	/* Check for V2 type 4 tag */
-	err = ISO_Select(iso_appname_v2, ARRAY_SIZE(iso_appname_v2),
-			0x4, check_presence, cookie);
-	if (err < 0)
-		goto out_err;
-
-	return err;
-
-out_err:
-	return t4_cookie_release(err, cookie);
+	return ISO_Select(iso_appname_v2, ARRAY_SIZE(iso_appname_v2),
+				0x4, check_presence, cookie);
 }
 
 static int select_ndef_file(uint8_t *resp, int length, void *data)
 {
 	struct t4_cookie *cookie = data;
-	int err = 0;
 
 	DBG("%d", length);
 
-	if (length < 0) {
-		err = length;
-		goto out_err;
-	}
+	if (length < 0)
+		return t4_cookie_release(length, cookie);
 
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
 		near_error("select ndef file resp failed %02X",
 					resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	DBG("ndef file selected");
@@ -854,8 +770,7 @@ static int select_ndef_file(uint8_t *resp, int length, void *data)
 	if (cookie->cb)
 		cookie->cb(cookie->adapter_idx, cookie->target_idx, 0);
 
-out_err:
-	return t4_cookie_release(err, cookie);
+	return t4_cookie_release(0, cookie);
 }
 
 static int read_cc_file(uint8_t *resp, int length, void *data)
@@ -928,15 +843,15 @@ static int select_cc_file(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		near_error("CC file select resp failed %d", length);
-		err = length;
-		goto out_err;
+
+		return t4_cookie_release(length, cookie);
 	}
 
 	if (APDU_STATUS(resp + STATUS_WORD_1) != APDU_OK) {
 		near_error("CC file select response %02X",
 						resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	err = ISO_ReadBinary(0, LEN_ISO_CC_READ_SIZE, read_cc_file, cookie);
@@ -958,15 +873,15 @@ static int select_iso_appname_v2(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		near_error("iso app select resp failed %d", length);
-		err = length;
-		goto out_err;
+
+		return t4_cookie_release(length, cookie);
 	}
 
 	if (resp[NFC_STATUS] != 0x00) {
 		near_error("iso app select response %02X",
 					resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	err = ISO_Select(iso_cc_fileid, LEN_ISO_CC_FILEID, 0,
@@ -992,22 +907,20 @@ static int format_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		near_error("write data to ndef file resp failed %d", length);
-		err = length;
-		goto out_err;
+
+		return t4_cookie_release(length, cookie);
 	}
 
 	if (APDU_STATUS(resp + 1) != PICC_LEVEL_APDU_OK) {
 		near_error("wrtie data to ndef file response %02X",
 						resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
-	if (tag == NULL) {
-		err = -EINVAL;
-		goto out_err;
-	}
+	if (tag == NULL)
+		return t4_cookie_release(-EINVAL, cookie);
 
 	DBG("Formatting is done");
 	near_tag_set_blank(tag, FALSE);
@@ -1026,7 +939,7 @@ static int format_resp(uint8_t *resp, int length, void *data)
 		goto out_err;
 	}
 
-	return 0;
+	return err;
 
 out_err:
 	return t4_cookie_release(err, cookie);
@@ -1426,8 +1339,8 @@ static int get_version_frame3(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		near_error("get version2 resp failed %d", length);
-		err = length;
-		goto out_err;
+
+		return t4_cookie_release(length, cookie);
 	}
 
 	if (resp[4] == 0x01 /* Major Version */
@@ -1439,18 +1352,16 @@ static int get_version_frame3(uint8_t *resp, int length, void *data)
 					select_application, cookie);
 		if (err < 0) {
 			near_error("get version3 req failed %d", err);
-			goto out_err;
+
+			return t4_cookie_release(err, cookie);
 		}
 	} else {
 		near_error("get version2 response %02X", resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	return 0;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 static int get_version_frame2(uint8_t *resp, int length, void *data)
@@ -1462,8 +1373,8 @@ static int get_version_frame2(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		near_error(" get version resp failed %d", length);
-		err = length;
-		goto out_err;
+
+		return t4_cookie_release(length, cookie);
 	}
 
 	if (resp[4] == 0x01 /* Major Version */
@@ -1480,18 +1391,16 @@ static int get_version_frame2(uint8_t *resp, int length, void *data)
 					get_version_frame3, cookie);
 		if (err < 0) {
 			near_error("get version2 req failed %d", err);
-			goto out_err;
+
+			return t4_cookie_release(err, cookie);
 		}
 	} else {
 		near_error("get version response %02X", resp[length - 1]);
-		err = -EIO;
-		goto out_err;
+
+		return t4_cookie_release(-EIO, cookie);
 	}
 
 	return 0;
-
-out_err:
-	return t4_cookie_release(err, cookie);
 }
 
 /* Steps to format Type 4 (MIFARE DESFire EV1) tag as per AN1104.pdf from nxp.
